"""
$RCSfile: CVSSandboxRegistry.py,v $

The CVSFile package is a Zope product.  CVSFile allows one to use
CVS as a version control system for Zope content, by pointing to a
file in the developer's CVS sandbox on the Zope server.
CVSSandboxRegistry encapsulates information about CVS sandboxes, which is
obtained by CVSFile objects via acquisition.

Author: <a href="mailto:cstrong@arielpartners.com">Craeg Strong</a>
Release: 1.0
"""

__cvstag__  = '$Name:  $'[6:-2]
__date__    = '$Date: 2002/04/01 01:36:57 $'[6:-2]
__version__ = '$Revision: 1.6 $'[10:-2]

#
# Add our base classes.  Derived classes will get this stuff automatically
#
from OFS.SimpleItem import SimpleItem  # The SimpleItem module is a Zope builtin that provides
                                       # cut/paste, management views, webdav/ftp, undo, ownership,
                                       # traversal, acquisition, and persistence
from Products.ZCatalog.CatalogAwareness import CatalogAware # Reqd for ZCatalog support
from OFS.PropertyManager import PropertyManager  # Reqd for proper property mgmt
from SandboxInfo import SandboxInfo
from ISandboxRegistry import ISandboxRegistry

# Zope builtins
from Globals import DTMLFile, MessageDialog
from AccessControl import ClassSecurityInfo
import Globals  # InitializeClass (security stuff)
import OFS                   # we grab his _property settings
from DateTime.DateTime import DateTime
from webdav.common import rfc1123_date

# Python builtins
from cStringIO import StringIO
import errno, os, os.path, socket, string, sys

################################################################
# Utility functions
################################################################

def canonicalizeDir(dirname):
    """

    Adds a trailing slash to a directory, so they all look the same

    """
    lastChar = dirname[len(dirname) - 1]
    if lastChar != os.sep:
        return dirname + os.sep
    else:
        return dirname

command_separator       = ';'
if sys.platform == 'win32':
    command_separator = '&'

def runCommand(command):
    """

    send passed in commandline to OS, return output or '(no
    output)' if there is no output

    """
    #print "running this command", command
    input, output_and_error  = os.popen4(command)
    output = output_and_error.read()
    if not output:
        output = "(no output)"
    return output

################################################################
# Container Methods
################################################################

# globals() is standard python for passing the current namespace
manage_addCVSSandboxRegistryForm = DTMLFile('dtml/create_registry', globals())

def manage_addCVSSandboxRegistry(self,
                                 id='', title='', description='',                  # standard
                                 sandboxname='',sandboxdir='',cvsroot='',          # cvs registry
                                 REQUEST=None, RESPONSE=None):
    """
    
    Add a CVSSandboxRegistry to a folder.  Called from the
    create_registry.dtml form.

    """
    if not id or not sandboxname or not sandboxdir or not cvsroot:
        return MessageDialog(title   = 'ERROR',
                             message = 'Required fields must not be blank',
                             action  = 'manage_main')
    
    if not os.path.exists(sandboxdir) or not os.path.isdir(sandboxdir):
        return MessageDialog(title   = 'ERROR',
                             message = sandboxdir + ' is not a valid directory',
                             action  = 'manage_main')

    self._setObject(id, CVSSandboxRegistry(id, title,
                                           description,
                                           SandboxInfo(sandboxname,
                                                       canonicalizeDir(sandboxdir),
                                                       cvsroot) ))
    self._getOb(id).reindex_object()

    message = 'successfully created ' + id + ' registry object.'
    return MessageDialog(title   = 'CREATED',
                         message = message,
                         action  = 'manage_main')

################################################################
# CVSSandboxRegistry class
################################################################

# By default when doing a recursive checkout of the
# entire sandbox, checkout this module
default_cvs_module      = 'com'

# By default, whenever doing a CVS command, specify this argument
# -P is "prune empty directories"
default_cvs_args        = '-P'

# default cookie name
cookie_name             = 'CVSSandbox'

class CVSSandboxRegistry(CatalogAware,    # ZCatalog support
                         PropertyManager, # Property support
                         SimpleItem):

    """
    CVSSandboxRegistry
    """

    meta_type = 'CVS Sandbox Registry'  # This is the name Zope will use for the Product in
    # the "addProduct" list

    __implements__ = ISandboxRegistry
    
    # This tuple defines our properties
    _properties = (
        {'id':'title',                               'type':'string',    'mode': 'w'},
        {'id':'description',                         'type':'string',    'mode': 'w'},
        {'id':'default_cvs_module_to_checkout',      'type':'string',    'mode': 'w'},
        {'id':'default_cvs_command_line_parameters', 'type':'string',    'mode': 'w'},        
        )

    # This tuple defines a dictionary for each tab in the management interface
    # label = label of tab, action = url it links to
    manage_options = (
        {'label':'Registry', 'action': 'manage_registryForm', 'help': ('CVSFile', 'registry.stx') },
        {'label':'Sandbox',  'action': 'manage_sandboxForm',  'help': ('CVSFile', 'sandbox.stx') },
        ) + OFS.PropertyManager.PropertyManager.manage_options + OFS.SimpleItem.SimpleItem.manage_options

    _security = ClassSecurityInfo()

    # set security for the object itself, e.g. if it is accessed in DTML code
    # This line is REQUIRED to allow access to CVSFiles by the public (ariel DTDs, etc.)
    _security.declareObjectPublic()

    _security.declareProtected('Manage CVS Sandboxes', 'manage_sandboxForm')
    manage_sandboxForm = DTMLFile('dtml/sandbox',globals())

    _security.declareProtected('Manage CVS Sandboxes', 'manage_registryForm')
    manage_registryForm = DTMLFile('dtml/registry',globals())

    def __init__(self, id, title, description, default_sandbox):
        """
        Initialize a new instance of CVSSandboxRegistry
        """
        self.id             = id       # This is the OID translated by Zope into a URL
        self.title          = title
        self.description    = description
        self.__version__    = __version__
        self.sandboxes      = { default_sandbox.name : default_sandbox }
        self.defaultSandbox = default_sandbox
        
        self.default_cvs_module_to_checkout      = default_cvs_module
        self.default_cvs_command_line_parameters = default_cvs_args        

    ################################################################
    # Methods implementing the ISandboxRegistry interface
    ################################################################

    def sandboxList(self):
        "Returns the list of sandboxes"
        return self.sandboxes.values()
        
    def numSandboxes(self):
        "return the number of sandboxes currently registered"
        return len(self.sandboxes)

    def getSandboxByName(self, sandboxName):
        "return the named sandbox, or None"
        return self.sandboxes.get(sandboxName, None)

    _security.declareProtected('View',  'isSelected')
    def isSelected(self, sandboxName, REQUEST):
        """
        
        Return 1 if name represents the currently selected sandbox,
        or 0 if it is not, or if the name does not exist

        """
        return REQUEST and REQUEST.has_key(cookie_name) and REQUEST[cookie_name] == sandboxName

    _security.declareProtected('View',  'isDefault')
    def isDefault(self, sandboxName):
        """
        
        Return 1 if name represents the default sandbox,
        or 0 if it is not, or if the name does not exist

        """
        return self.defaultSandbox.name == sandboxName

    _security.declareProtected('View',  'getSelectedSandbox')
    def getSelectedSandbox(self, REQUEST):
        "return the selected sandbox or None"
        if REQUEST and REQUEST.has_key(cookie_name):
            return self.getSandboxByName(REQUEST[cookie_name])
        else:
            return None

    def getDefaultSandbox(self):
        "return the default sandbox"
        return self.defaultSandbox

    _security.declareProtected('View',  'getCurrentSandbox')
    def getCurrentSandbox(self, REQUEST):
        """

        return the selected sandbox or the default, if none has been
        selected

        """
        si = self.getSelectedSandbox(REQUEST)
        if si:
            return si
        else:
            return self.getDefaultSandbox()

    def setDefaultSandbox(self, sandboxName):
        """

        set the default sandbox to the one whose name is passed.  Barf
        if the name does not exist in this registry

        """
        if not self.sandboxes.has_key(sandboxName):
            message = 'No sandbox named ' + sandboxName + ' exists in this registry.'
            raise ValueError, message
        si = self.sandboxes[sandboxName]
        self.defaultSandbox = si
        self._p_changed = 1  # dirty flag, triggers persistence to update the object        

    _security.declareProtected('View',  'selectSandbox')
    def selectSandbox(self, sandboxName, REQUEST):
        """
        
        Select the named sandbox.  Barf if the named sandbox does not
        exist.

        """
        if not self.sandboxes.has_key(sandboxName):
            message = 'No sandbox named ' + sandboxName + ' exists in this registry.'
            raise ValueError, message
        if REQUEST:
            e=(DateTime('GMT') + 365).rfc822()
            resp = REQUEST['RESPONSE']
            #
            # path   = "/" means apply to all pages within this website
            # expire = "+1yr means cookie stays around for a year"
            #
            # see www.netscape.com/newsref/std/cookie_spec.html
            #
            resp.setCookie(cookie_name, sandboxName,
                           path='/', expires=e)

    def addSandbox(self, si):
        """
        
        Add a sandbox to the registry.  Barf if it has the same name
        as an existing sandbox

        """
        if self.sandboxes.has_key(si.name):
            message = 'A sandbox named ' + si.name + ' already exists in this registry.'
            raise ValueError, message
        self.sandboxes[si.name] = si
        self._p_changed = 1  # dirty flag, triggers persistence to update the object        

    def removeSandbox(self, sandboxName):
        """
        
        Remove a sandbox from the registry.  Barf if it does not exist
        or is the default sandbox

        """
        if not self.sandboxes.has_key(sandboxName):
            message = 'No sandbox named ' + sandboxName + ' exists in this registry.'
            raise ValueError, message

        if self.defaultSandbox.name == sandboxName:
            message = 'The default sandbox can not be removed.'
            raise ValueError, message

        del self.sandboxes[sandboxName]
        self._p_changed = 1  # dirty flag, triggers persistence to update the object        

    _security.declareProtected('Manage CVS Sandboxes', 'changeSandbox')
    def changeSandbox(self, nameOfSandboxToChange, newSI):
        """
        
        Change the named sandbox to have the new name, dir, and
        cvsroot values.

        """
        if not self.sandboxes.has_key(nameOfSandboxToChange):
            message = 'No sandbox named ' + nameOfSandboxToChange + ' exists in this registry.'
            raise ValueError, message

        if self.sandboxes.has_key(newSI.name):
            message = 'A sandbox named ' + newSI.name + ' already exists in this registry.'
            raise ValueError, message
            
        si = self.sandboxes[nameOfSandboxToChange]
        si.name    = newSI.name
        si.basedir = newSI.basedir
        si.cvsroot = newSI.cvsroot
        del self.sandboxes[nameOfSandboxToChange]
        self.sandboxes[newSI.name] = si
        self._p_changed = 1  # dirty flag, triggers persistence to update the object

    def checkoutSandbox(self, checkoutModule, checkoutArgs, si):
        """
        
        Does a CVS checkout recursively on the entire sandbox, using
        the sandbox information from the SandboxInfo parameter

        """
        command = "cd %s %s cvs -d %s -q checkout %s %s" % ( si.basedir,
                                                             command_separator,
                                                             si.cvsroot,
                                                             checkoutArgs,
                                                             checkoutModule )
        output = runCommand(command)
        return command + "\n" + output

    ################################################################
    # Methods called from sandbox form
    ################################################################

    # methods called from forms
    _security.declareProtected('Manage CVS Sandboxes',  'manage_addSandbox')
    def manage_addSandbox(self,
                          name='', basedir='', cvsroot='',
                          REQUEST=None, RESPONSE=None):
        """
        
        Add a new sandbox to the registry.  Called from
        registry.dtml form.
            
        """
        message = None
        if not name or not basedir or not cvsroot:
            message = 'All values must be filled in'
        elif not os.path.exists(basedir) or not os.path.isdir(basedir):
            message = basedir + ' is not a valid SANDBOXDIR'
            
        if message:
            if REQUEST:
                return MessageDialog(title   = 'ERROR',
                                     message = message,
                                     action  = 'manage_registryForm')
            else:
                raise ValueError, message
        try:
            si = SandboxInfo(name, canonicalizeDir(basedir), cvsroot)
            self.addSandbox(si)
        except ValueError, message:
            return MessageDialog(title   = 'ERROR',
                                 message = message,
                                 action  = 'manage_registryForm')
        return RESPONSE.redirect('../manage_main')

    _security.declareProtected('Manage CVS Sandboxes',  'manage_selectSandbox')
    def manage_selectSandbox(self, boxes=None,
                             REQUEST=None, RESPONSE=None):
        """
        
        Select a sandbox to the registry.  Called from registry.dtml
        form.  More than one cannot be selected, b/c we are using a
        radio box for selection.
            
        """
        if not boxes:
            return MessageDialog(title   = 'ERROR',
                                 message = 'You must first select a sandbox.',
                                 action  = 'manage_registryForm')
        try:
            for struct in boxes:
                if struct.has_key('checked'):
                    self.selectSandbox(struct['checked'], REQUEST)
        except ValueError, message:
            return MessageDialog(title   = 'ERROR',
                                 message = message,
                                 action  = 'manage_registryForm')
        return RESPONSE.redirect('../manage_main')

    _security.declareProtected('Manage CVS Sandboxes',  'manage_removeSandbox')
    def manage_removeSandbox(self, boxes=None,
                             REQUEST=None, RESPONSE=None):
        """
        
        Remove a sandbox from the registry.  Called from registry.dtml
        form.  More than one cannot be removed at one time, b/c we are
        using a radio box for selection.
            
        """
        if not boxes:
            return MessageDialog(title   = 'ERROR',
                                 message = 'You must first select a sandbox.',
                                 action  = 'manage_registryForm')
        try:
            for struct in boxes:
                if struct.has_key('checked'):
                    self.removeSandbox(struct['checked'])
        except ValueError, message:
            return MessageDialog(title   = 'ERROR',
                                 message = message,
                                 action  = 'manage_registryForm')
        return RESPONSE.redirect('../manage_main')

    _security.declareProtected('Manage CVS Sandboxes',  'manage_setDefaultSandbox')
    def manage_setDefaultSandbox(self, boxes=None,
                                 REQUEST=None, RESPONSE=None):
        """
        
        Designate a particular sandbox in the registry as the default.
        Called from registry.dtml form.  More than one cannot be
        selected, b/c we are using a radio box for selection.
            
        """
        if not boxes:
            return MessageDialog(title   = 'ERROR',
                                 message = 'You must first select a sandbox.',
                                 action  = 'manage_registryForm')
        try:
            for struct in boxes:
                if struct.has_key('checked'):
                    self.setDefaultSandbox(struct['checked'])
        except ValueError, message:
            return MessageDialog(title   = 'ERROR',
                                 message = message,
                                 action  = 'manage_registryForm')
        return RESPONSE.redirect('../manage_main')

    _security.declareProtected('Manage CVS Sandboxes',  'manage_setDefaultSandbox')
    def manage_changeSandbox(self, boxes,
                             REQUEST=None, RESPONSE=None):
        """
        
        Change the named sandbox to have the new name, dir, and
        cvsroot values.
            
        """
        return MessageDialog(title   = 'ERROR',
                             message = 'Not implemented yet',
                             action  = 'manage_registryForm')

    # called from sandbox.dtml form
    _security.declareProtected('Manage CVS Sandboxes',  'cvsCheckoutSandbox')
    def manage_CheckoutSandbox(self, checkoutModule='', checkoutArgs='',
                               REQUEST=None):
        """
        
        Does a CVS checkout recursively on the entire sandbox and
        passes output back to the form
        

        """
        si     = self.getCurrentSandbox(REQUEST)
        output = self.checkoutSandbox(checkoutModule, checkoutArgs, si)
        return self.manage_sandboxForm(checkoutSandboxOutput = output)

    ################################################################
    # Standard Zope stuff
    ################################################################

    _security.declareProtected('View',  '__str__')
    def __str__(self):
        message = self.id + " (" + self.title + ")"
        return message

    ################################################################
    # Support for Schema Migration
    ################################################################

    def repair(self):
        """

        Repair this object.  This method is used for schema migration,
        when the class definition changes and existing instances of
        previous versions must be updated.

        """

# register security information
Globals.InitializeClass(CVSSandboxRegistry)
